/*******************************************************************************

NAME            nubd.h
UPDATE DATE     %date_modified: Thu Jun 08 10:09:07 2017 %
PROGRAMMER      %created_by:    mben %

    Copyright 2017-2018 NetApp, Inc.

DESCRIPTION:
    Definitions for the NVMe user space kernel driver.

LICENSE:
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*******************************************************************************/

#ifndef _UAPI_LINUX_NUBD_H
#define _UAPI_LINUX_NUBD_H

#include <linux/types.h>

struct nubd_ctrl_config
{
    int      handle;         // Set by kernel driver
    int      max_cores;      // Hint from application indicating max number of
                             // cores it expects to be used.  This will help the kernel
                             // driver align admin and IO queue resources appropriately,
                             // especially using the controller memory buffer.
                             // TODO: consider using cpu set to both define number and which cores to use.
    void *   cmb_base;       // "Controller memory buffer" provide by application.
    size_t   cmb_size;       // "Controller memory buffer" provide by application.  Size should include
                             // room for queue pairs and PRPLs for max number of queues supported.
    void *   reg_base;       // pseudo-register memory area, including SQ and CQ doorbells
    size_t   reg_size;       // pseudo-register memory area size
};

// NOTE: Currently the ioctl only support a single SGL
//       segment including with all data blocks to transfer.
typedef struct nubd_sgl_desc
{
    __u64    addr;
    __u32    length;
    __u8     reserved[3];
    struct {
        __u8 zero    : 4;
        __u8 type    : 4;
    } flags;
} NubdSglDesc;

#define NUBD_SGL_DATABLOCK_DESC  0
#define NUBD_SGL_LAST_SEG_DESC   3

// NOTE to maintainers:  Usage of this structure may include aligned structure pools,
// so keeping the structure size to no more than 64 bytes would be generally good.
struct nubd_rw_data_copy_req
{
    __u64   prp1;     // HW address of first DMA address (4-byte aligned)
    __u64   prp2;     // HW address of second page when only 2 pages.
                      // HW address of PRP list when more than 2 pages.
    struct  nubd_sgl_desc srv_sgl; // Segment descriptor for first SGL segment.
                      // Addresses within SGL should be user space addresses.
    __u32   offset;   // Byte offset into the client area given by prp1/2
    __u32   prp_data_size;  // Total length of client request.
    size_t   srv_sgl_nents;
    size_t   xfer_len;  // Bytes to transfer in this copy request.
    __u16   pi_flags;   // Indicates what DA actions need to be taken in copy.
    __u16   pi_apptag;  // Apptag for all PI blocks, cpu order
    __u32   pi_reftag;  // Ref tag for first PI block, cpu order. assume type 1: Incremented for each block processed.
};

/* Bit flag values for pr_flags field in the nubd_rw_data_copy_req structure. */
#define NUBD_PI_NO_PI           0x0000
#define NUBD_PI_HAS_PI          0x0001
#define NUBD_PI_CHECK_APPTAG    0x0002      // Only relevant on reads.
#define NUBD_PI_CHECK_REFTAG    0x0004      // Only relevant on reads.
#define NUBD_PI_CHECK_ALL       NUBD_PI_HAS_PI | NUBD_PI_CHECK_APPTAG | NUBD_PI_CHECK_REFTAG

struct nubd_io_mem_range
{
    __u64   p_addr;
    size_t  length;
};

// MGMT IOCTLs
#define NUBD_IOCTL_CREATE_DEV	    _IOWR('N', 0x50, struct nubd_ctrl_config)
#define NUBD_IOCTL_DESTROY_DEV	    _IO('N', 0x51)

// CTRL IOCTLS
#define NUBD_IOCTL_RD_COMPLETE	    _IOW('N', 0x58, struct nubd_rw_data_copy_req)
#define NUBD_IOCTL_WR_COMPLETE	    _IOW('N', 0x59, struct nubd_rw_data_copy_req)
#define NUBD_IOCTL_START_CTRL	    _IO('N', 0x5a)  // Ready to accept IO.
#define NUBD_IOCTL_CQ_TAIL_UPDATE	_IO('N', 0x5b)
#define NUBD_IOCTL_DISK_UPDATE	    _IO('N', 0x5c)
#define NUBD_IOCTL_RESCAN_DISKS     _IO('N', 0x5d)
#define NUBD_IOCTL_ADD_IOMEM        _IOWR('N', 0x5e, struct nubd_io_mem_range)

// Disk IOCTLS
#define NUBD_IOCTL_NSID	        _IO('N', 0x60)

// Related device name constants
#define NUBD_MGR_DEV    "/dev/nubd"

#endif /* _UAPI_LINUX_NUBD_H */
